#ifndef IMAGE_FILEOUTPUT_H
#define IMAGE_FILEOUTPUT_H

#include <iostream>
#include <list>
#include <Magick++.h>
#include <map>
#include "image_output.h"
#include "rgba.h"

namespace mimas {

/** @addtogroup imageIO
    @{ */
/** @addtogroup imageOutput
    @{ */
/** Sink for writing an image to a file.
    The stream is used for writing image(s) to. The stream-manipulator
    \c setformat can be used to change the file-format.

    The <A HREF="http://www.imagemagick.org/Magick++/">Magick++</A> library
    is used for writing the images.

    Currently <A HREF="http://www.imagemagick.org/Magick++/">Magick++</A>
    supports writing the following file-formats: 8BIM 8BIMTEXT 8BIMWTEXT APP1
    APP1JPEG AVS BIE BMP BMP2 BMP3 CMYK CMYKA DCX DPX EPDF EPI EPS EPS2 EPS3
    EPSF EPSI EPT EXIF FAX FITS FPX G3 GIF GIF87 GRAY HISTOGRAM HTM HTML ICB
    ICC ICM IPTC IPTCTEXT IPTCWTEXT JBG JBIG JNG JPEG JPG LOGO M2V MAP MATTE
    MIFF MNG MONO MPC MPEG MPG MSL MTV MVG NULL OTB P7 PAL PALM PBM PCD PCDS
    PCL PCT PCX PDB PDF PGM PICON PICT PM PNG PNG24 PNG32 PNG8 PNM PPM PREVIEW
    PS PS2 PS3 PSD PTIF RAS RGB RGBA SGI SHTML SUN SVG TEXT TGA TIF TIFF TXT
    UIL UYVY VDA VICAR VID VIFF VST WBMP X XBM XMP XPM XV XWD YUV

    @author Jan Wedekind (jan@wedesoft.de)
    @date Thu Jan 12 14:19:44 UTC 2006
    @see image_fileinput
    @see setformat */
template< typename T >
class image_fileoutput: public image_output< T >
{
 public:
  /** Constructor.
      @param _stream Stream to write images to. */
  explicit image_fileoutput( std::ostream &_stream ): stream( &_stream ) {}
  /** Destructor.
      The destructor will write the images to the stream! */
  virtual ~image_fileoutput(void);
  /** Write image to stream.
      @param img Image to write to the stream. */
  virtual void write( const const_image_ref< T > &image );
  /** Get list of supported file-formats.
      Gets a map of file-formats, which have write-support by Magick++. The
      boolean is indicating, wether the file-format allows saving of multiple
      images or not. */
  static std::map< std::string, bool > formats(void);
 protected:
  /** Identification of mapping.
      This method returns a parameter passed to \c Magick::Image::write. The
      return value either is "BGRA" or "I". */
  static const char *mapString(void);
  /** Identification of default output format.
      This method returns a parameter passed to \c Magick::Image::magick. The
      return value either is "PPM" or "PGM". */
  static const char *defaultFormat(void);
  /// Pointer to stream.
  std::ostream *stream;
  /// Buffer for multiple images.
  std::list< Magick::Image > buffer;
};

template<>
const char *image_fileoutput< unsigned char >::mapString(void);

template<>
const char *image_fileoutput< rgba< unsigned char > >::mapString(void);

template<>
const char *image_fileoutput< unsigned char >::defaultFormat(void);

template<>
const char *image_fileoutput< rgba< unsigned char > >::defaultFormat(void);

/// File-format of next image to be written.
extern const char *image_fileoutput_fileType;

///
struct _format { const char *name; };
/** Stream-manipulator to change the file-format.
    The Qt-documentation contains information about what file formats are
    supported (class \c QImageIO). After the image was written, the default
    value will be restored.
    @see image_fileoutput
    @see image_fileoutput_fileType */
inline _format setformat( const char *name )
{ _format retVal; retVal.name = name; return retVal; }

/** Change the file-format for the next image.
    Unfortunately this method is not side-effect free.
    @see image_fileoutput_fileType */
inline std::ostream &operator<<( std::ostream &stream, _format format )
{ image_fileoutput_fileType = format.name; return stream; }

/// Stream-operator for writing image to a stream.
template< typename T, typename U >
inline image_fileoutput< T > &operator<<( image_fileoutput< T > &output,
                                          const const_image_ref< U > &im )
{
  if ( typeid( T ) == typeid( U ) )
    output.write( (const_image_ref< T > &)im );
  else {
    output.write( image< T >( im ) );
  };
  return output;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< unsigned char > &img )
{
  image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< int > &img )
{
  image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< double > &img )
{
  image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< float > &img )
{
  image_fileoutput< unsigned char > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< rgba< unsigned char > > &img )
{
  image_fileoutput< rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< rgba< int > > &img )
{
  image_fileoutput< rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< rgba< float > > &img )
{
  image_fileoutput< rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

/// For convencience
inline std::ostream &operator<<( std::ostream &stream,
                                 const const_image_ref< rgba< double > > &img )
{
  image_fileoutput< rgba< unsigned char > > output( stream );
  output << img;
  return stream;
}

///@}

///@}

};

#include "image_fileoutput.tcc"

#endif
