#ifndef MESA_BASE_H
#define MESA_BASE_H

#include "mimasconfig.h"

#ifdef HAVE_OPENGL

#include <GL/gl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "mimasexception.h"
#include "object.h"
#include "rgba.h"
#include "x11_display.h"

namespace mimas {

/** Abstract base class for mesa-functionality.
    The mesa-functionality mainly consists in offering initialisation of
    an OpenGL-context for on-screen- or off-screen-rendering.

    After initialisation of a deriving non-abstract class, the OpenGL
    functionality can be either accessed directly using the OpenGL methods
    bypassing this class or accessed indirectly by using one of the few
    wrappers for setting the color, for drawing rectangles
    and for displaying text, being provided by this class.
    
    At any time there must be only one instance of this class to avoid
    crashes of X11, side-effects or deadlocks.

    Xorg 6.9.0 does not display gray images properly (GL_LUMINANCE). A bug
    report was filed:
    http://bugs.freedesktop.org/show_bug.cgi?id=7444

    If you experience problems, you may want to try:
    \verbatim
    export LIBGL_ALWAYS_INDIRECT=1
    \endverbatim
    This disables hardware rendering.

    If it works:
    - the problem is likely with the display driver
    - it will be very slow

    If it doesn't:
    - there is a different bug, we need to fix

    Do not forget to initialise <EM>glut</EM> at the beginning of your
    program:
    \verbatim
    glutInit( &argc, argv );
    \endverbatim

    References:
    \li <A HREF="http://www.gamedev.net/download/redbook.pdf">http://www.gamedev.net/download/redbook.pdf</A>
    @author Jan Wedekind (jan@wedesoft.de)
    @date Wed Mar 30 19:35:05 UTC 2005 */
class mesa_base: public object
{
 public:
  /** Destructor.
      The destructor just calls the method \c finish() .
      @see finish */
  virtual ~mesa_base(void);
  /** Finish output on the image.
      This method allows finalising of the output without having to destruct
      this object immediately.

      Only call after having finished rendering. This object has to be
      destructed afterwards. */
  virtual void finish(void);
  /** Switch to two-dimensional drawing mode.
      The coordinate system has the origin in the "top-left" corner of the
      image. */
  virtual void ortho2D(void) const = 0;
  /** Switch to two-dimensional drawing mode with clipping.
      The coordinate system has the origin in the "top-left" corner of the
      image. */
  virtual void ortho2D( int x1, int y1, int x2, int y2 ) const = 0;
  /** Set color.
      @param color Color-object. The alpha-component is ignored. */
  void setColor( const rgba< unsigned char > &color ) const
  { glColor3ub( color.getRed(), color.getGreen(), color.getBlue() ); }
  /** Set luminance.
      @param gray The gray-value. */
  void setGray( unsigned char gray ) const
  { glColor3ub( gray, gray, gray ); }
  /** Draw rectangle.
      @param x1 x-coordinate of upper-left corner of rectangle.
      @param y1 y-coordinate of upper-left corner of rectangle.
      @param x2 x-coordinate of lower-right corner of rectangle.
      @param y2 y-coordinate of lower-right corner of rectangle. */
  void rectangle( int x1, int y1, int x2, int y2 ) const;
  /// Write text.
  void text( const std::string &s ) const;   
 protected:
  /** Protected constructor.
      The \c display and \c visualInfo variables must be destructed by the
      calling program.
      @param _display X11 display variable.
      @param _visualInfo X11 visual info structure. */
  mesa_base( x11_display *_display, XVisualInfo *_visualInfo )
    throw (mimasexception);
  /** Wrapper for converting OpenGL error messages to exceptions.
      The wrapper will generate an exception (using \c gluErrorString), if
      the condition is not fullfilled.
      @param condition If the condition is \c false, an exception is thrown.
      @param message Textual description of context, which is prepended to
      the error-message. */
  void glWrap( bool condition, const char *message ) throw (mimasexception);
  /// X11-display.
  x11_display *display;
  /// X11-visual with desired attributes.
  XVisualInfo *visualInfo;
  /// State of object.
  bool finishBase;
  /// There must be only one instance of this class.
  static bool existInstance;
};

};

#endif

#endif

