#ifndef MESA_OFFSCREEN_TCC
#define MESA_OFFSCREEN_TCC

namespace mimas {

template< typename T >
mesa_offscreen< T >::mesa_offscreen( x11_display *_display,
				     image< T > &image,
				     bool depthBuffer ):
  mesa_base( _display, NULL ), width( image.getWidth() ),
  height( image.getHeight() ), finishOffscreen( false )
{
#ifndef HAVE_OSMESA
  // #warning "Recent code for off-screen rendering with X11-pixmaps hasn't been tested."
  // Use X11-pixmap for off-screen rendering (still requires an X11-server).

  // Request desired X11 visual.
  int attribs[] = { GLX_RGBA,
                    GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1,
                    GLX_DEPTH_SIZE, depthBuffer ? 16 : 0, None };
  visualInfo = glXChooseVisual( display->get(),
                                DefaultScreen( display->get() ),
                                attribs );
  glWrap( visualInfo != NULL, "Error choosing desired X11 visual" );

  try {

    // Create an OpenGL-context.
    context = glXCreateContext( display->get(), visualInfo, NULL, False );

    glWrap( context != NULL, "Error creating GLX context" );

    try {

      // Create an X11-pixmap for off-screen rendering.
      pixmap = XCreatePixmap( display->get(),
                              DefaultRootWindow( display->get() ),
                              width, height, visualInfo->depth );
      MMERROR( pixmap != 0, mimasexception, , "Couldn't create X11 pixmap." );

      try {

        // Attach OpenGL-pixmap to X11-pixmap.
        glxPixmap = glXCreateGLXPixmap( display->get(), visualInfo, pixmap );

        glWrap( glxPixmap != 0, "Couldn't create GLX pixmap" );

        // Perform context-switch.
        if ( glXMakeCurrent( display->get(), glxPixmap, context ) != True ) {
            
          glXDestroyGLXPixmap( display->get(), glxPixmap );
          MMERROR( false, mimasexception, , "Failed switching GLX context." );

        };

      } catch ( mimasexception &e ) {

        XFreePixmap( display->get(), pixmap );
        throw e;

      };
    } catch ( mimasexception &e ) {

      glXDestroyContext( display->get(), context );
      throw e;

    };      
  } catch ( mimasexception &e ) {

    // XFree( visualInfo );
    throw e;

  };

  data = image.rawData();
  
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
  // Direct access to image-memory!
  glDrawPixels( width, height, getFormat(), GL_UNSIGNED_BYTE, data );
      
#else
  // Use OSMesa-API for off-screen rendering. Initialisation of OSMesa is
  // a bit easier, but most X-Servers are not supporting it at the moment.

  // Create OSMesa-context.
  context = OSMesaCreateContextExt( getFormat(), depthBuffer ? 16 : 0, 0, 0, NULL );

  // Perform context-switch.
  // Direct access to image-memory!
  OSMesaMakeCurrent( context, image.rawData().data(), GL_UNSIGNED_BYTE,
                     width, height );

#endif
}

template< typename T >
mesa_offscreen< T >::~mesa_offscreen(void)
{
  // Finish off-screen rendering.
  finish();    
}

template< typename T >
void mesa_offscreen< T >::finish(void)
{

  if ( !finishOffscreen ) {

    glFinish();

#ifndef HAVE_OSMESA
    // Copy back image-data.
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );
    // Direct access to image-memory!
    setupColorScale();
    glReadPixels( 0, 0, width, height, getFormat(), GL_UNSIGNED_BYTE, data );
    glPixelTransferf( GL_RED_SCALE, 1.0 );
    glPixelTransferf( GL_GREEN_SCALE, 1.0 );
    glPixelTransferf( GL_BLUE_SCALE, 1.0 );
    
    // Destroy all OpenGL- and X11-objects.
    glXDestroyGLXPixmap( display->get(), glxPixmap );
    XFreePixmap( display->get(), pixmap );
    glXDestroyContext( display->get(), context );
    XFree( visualInfo );
#else
    // Destroy OSMesa-context.
    OSMesaDestroyContext( context );
#endif

    finishOffscreen = true;

  };


  mesa_base::finish();

}

template< typename T >
void mesa_offscreen< T >::ortho2D(void) const
{
  ortho2D( 0, 0, width, height );
}

template< typename T >
void mesa_offscreen< T >::ortho2D( int x1, int y1, int x2, int y2 ) const
{
  assert( !finishOffscreen );
  glLoadIdentity();
  glViewport( x1, y1, x2 + 1 - x1, y2 + 1 - y1 );
  glOrtho( x1, x2, y1, y2, -1, 1 );
}

template<>
GLenum mesa_offscreen< rgba< unsigned char > >::getFormat(void) const
  throw (mimasexception);

template<>
GLenum mesa_offscreen< unsigned char >::getFormat(void) const
  throw (mimasexception);

template<>
void mesa_offscreen< rgba< unsigned char > >::setupColorScale(void) const;

template<>
void mesa_offscreen< unsigned char >::setupColorScale(void) const;


};

#endif
