/**
 * Object recognition using chamfer matching
 * This class performs object recognition by using chamfer maps.
 * @author Bala Amavasai (bala@amavasai.org)
 * @date Mon Dec  3 19:28:39 GMT 2001
 * $Header: /cvs/mimas2/lib/chamfer_match.cc,v 1.1.1.1 2005/08/09 15:37:29 engmb Exp $
 */

#include <cstdlib>
#include <fstream>
#include "chamfer_match.h"
#include "image.h"
#include "mimasexception.h"
#include "image_funcs.h"
#include "xml_document.h"

namespace mimas {

using namespace std;

chamfer_match::chamfer_match(void)
{
    ang1=0.0;
    ang2=0.0;
    angstep=0.0;
    mlevels=1;
    chamferModel=true;
    sceneAdded=false;

    mincorr= FLT_MAX;

    roi_lx=0;
    roi_ly=0;
    roi_ux=0;
    roi_uy=0;
    ROI=false;

    cannyStdDev=1.0;
    cannyLowThres=31;
    cannyHighThres=127;
    cannyAutoThres=true;

    track_roi_lx=0;
    track_roi_ly=0;
    track_roi_ux=0;
    track_roi_uy=0;

    name = "undefined";
}

string	chamfer_match::getName(void)
{
    return name;
}

void	chamfer_match::setName( const std::string &n )
{
    name = n;
}

bool chamfer_match::getUseROI(void)
{
    return ROI;
}

void chamfer_match::useROI(bool val)
{
    ROI=val;
}

void chamfer_match::setROI(int x1, int y1, int x2, int y2)
{
    if (y1<y2)
    {
        roi_ly=y1;
        roi_uy=y2;
    }
    else
    {
        roi_ly=y2;
        roi_uy=y1;
    }

    if (x1<x2)
    {
        roi_lx=x1;
        roi_ux=x2;
    }
    else
    {
        roi_lx=x2;
        roi_ux=x1;
    }
}

void chamfer_match::getSearchAngleRange(float &angle1, float &angle2, float &anglestep)
{
    angle1=ang1;
    angle2=ang2;
    anglestep=angstep;
}

void chamfer_match::setSearchAngleRange(float angle1, float angle2, float anglestep, angle_type ang)
{
    if (angle1<angle2)
    {
        ang1=angle1;
        ang2=angle2;
    }
    else
    {
        ang1=angle2;
        ang2=angle1;
    }

    angstep=anglestep;

    // convert to radians
    if (ang==deg)
    {
        ang1=ang1*M_PI/180.0;
        ang2=ang2*M_PI/180.0;
        angstep=angstep*M_PI/180.0;
    }
}

void chamfer_match::setMultiResLevels(int levels)
{
    mlevels=levels;
}

void chamfer_match::setChamferModel(bool val)
{
    chamferModel=val;
}

bool chamfer_match::getChamferModel(void)
{
    return chamferModel;
}

int chamfer_match::getMultiResLevels(void)
{
    return mlevels;
}

void chamfer_match::setCannyAutoThres(bool val)
{
    cannyAutoThres=val;
}

void chamfer_match::setCannyStdDev(float val)
{
    cannyStdDev=val;
}

void chamfer_match::setCannyLowThres(int val)
{
    cannyLowThres=val;
}

void chamfer_match::setCannyHighThres(int val)
{
    cannyHighThres=val;
}

bool chamfer_match::getCannyAutoThres(void)
{
    return cannyAutoThres;
}

float chamfer_match::getCannyStdDev(void)
{
    return cannyStdDev;
}

int chamfer_match::getCannyLowThres(void)
{
    return cannyLowThres;
}

int chamfer_match::getCannyHighThres(void)
{
    return cannyHighThres;
}

float chamfer_match::getMinCorrVal(void)
{
    return mincorr;
}

void chamfer_match::addScene(image<int> &sceneimage)
{
    canny.setCannyAutoThres(cannyAutoThres);
    canny.setCannyStdDev(cannyStdDev);
    canny.setCannyLowThres(cannyLowThres);
    canny.setCannyHighThres(cannyHighThres);

    canny.getEdges(sceneimage,scene);

    sceneAdded=true;
}


void chamfer_match::findModel(image<int> &modelimage, int &est_x, int &est_y, float &est_anglerad) throw(mimasexception)
{
    MMERROR( sceneAdded, mimasexception, ,
             "chamfer_match::findModel - scene image not yet added!" );

    image<int> rotModel;

    model = modelimage;

    canny.getEdges(model);

    int x,y;
    float retcorr, angle;

    mincorr=FLT_MAX;
    est_anglerad=0.0;

    for (angle=ang1; angle<=ang2; angle+=angstep)
    {
        paethRotateRadian(model,rotModel,angle);

        multiLevelCorrelation(scene, rotModel, mlevels, x, y, retcorr);
        if (retcorr<mincorr)
        {
            mincorr=retcorr;
            est_x=x; est_y=y;
            est_anglerad=angle;
        }
    }

    est_x += modelimage.getWidth()/2;
    est_y += modelimage.getHeight()/2;
}

void chamfer_match::invert(image<int> &inimage)
{
    for (int j=0; j<inimage.getHeight(); j++)
        for (int i=0; i<inimage.getWidth(); i++)
          inimage.pixel(i,j)=255-inimage.pixel(i,j);
}


void chamfer_match::chamfer(image<int> &inimage)
{
    int nw,n,ne,w,e,sw,s,se;
    int fp;

    for (int j=1; j<inimage.getHeight()-1; j++)
        for (int i=1; i<inimage.getWidth(); i++)
        {
            fp=inimage.pixel(i,j);
            nw=4+inimage.pixel(i-1,j-1);
            n=3+inimage.pixel(i,j-1);
            w=3+inimage.pixel(i-1,j);
            sw=4+inimage.pixel(i-1,j+1);

            fp=fp<nw?fp:nw;
            fp=fp<n?fp:n;
            fp=fp<w?fp:w;
            fp=fp<sw?fp:sw;
            inimage.pixel(i,j) = fp;

        }

    for (int j=inimage.getHeight()-2; j>0; j--)
        for (int i=inimage.getWidth()-2; i>=0; i--)
        {
            fp=inimage.pixel(i,j);
            ne=4+inimage.pixel(i+1,j-1);
            e=3+inimage.pixel(i+1,j);
            s=3+inimage.pixel(i,j+1);
            se=4+inimage.pixel(i+1,j+1);

            fp=fp<ne?fp:ne;
            fp=fp<e?fp:e;
            fp=fp<s?fp:s;
            fp=fp<se?fp:se;
            inimage.pixel(i,j) = fp;

        }

    for (int i=0; i<inimage.getWidth(); i++)
    {
      inimage.pixel(i,0) = inimage.pixel(i,1);
      inimage.pixel(i,inimage.getHeight()-1) = inimage.pixel(i,inimage.getHeight()-2);
    }

    for (int j=0; j<inimage.getHeight(); j++)
    {
      inimage.pixel(0,j) = inimage.pixel(1,j);
      inimage.pixel(inimage.getWidth()-1,j) = inimage.pixel(inimage.getWidth()-2,j);
    }
}


void chamfer_match::paethRotateRadian(image<int> &imagein, image<int> &imageout, float radian)
{
    imageout.init(imagein.getWidth(),imagein.getHeight());
    // set background colour of object to be white
    imageout.fill(255);

    if (fabs(radian-M_PI)<1e-3) radian=radian<M_PI?M_PI-1e-3:M_PI+1e-3; // Quick hack to fix the problem when tan goes to infinity -bpa
    float s1=-tan(-radian/2.0);
    float s2=sin(-radian);

    float xr, yr;
    int x,y;
    for (int j=0; j<imageout.getHeight(); j++)
        for (int i=0; i<imageout.getWidth(); i++)
        {
            x=i-imageout.getWidth()/2;
            y=j-imageout.getHeight()/2;
            //x=i;
            //y=j;

            xr = (float)x + s1 * (float)y;
            yr = (float)y + s2 * xr;
            xr = xr + s1 * yr;

            xr+=(float)imageout.getWidth()/2;
            yr+=(float)imageout.getHeight()/2;
            imageout.pixel(i, j) = (int)imagein.getPixel((int)xr,(int)yr);
            //imageout.setPixel(i,j, imagein.getPixel(i,j));
        }
}

void chamfer_match::multiLevelCorrelation(image<int> &edgemap, image<int> &objectedgemap, int
        levels, int &x, int &y, float &corrval)
{
    image<int> object[levels], image[levels];

    image[0] = edgemap;
    // invert(image[0]);

    object[0] = objectedgemap;
    // invert(object[0]);

    for (int l=1; l<levels; l++)
    {
        halfResolutionEdgePyramid(image[l-1],image[l]);
        halfResolutionEdgePyramid(object[l-1],object[l]);
    }

    for (int l=0; l<levels; l++)
    {
        invert(image[l]);
        invert(object[l]);
        //image[l].display();
        //object[l].display();
        chamfer(image[l]);
        if (chamferModel) chamfer(object[l]);
    }

    int xmin, ymin;
    int xminsearch, xmaxsearch, yminsearch, ymaxsearch;
    float cmincorr;

    x=0; y=0;
    xminsearch=0;
    yminsearch=0;
    xmaxsearch=image[levels-1].getWidth();
    ymaxsearch=image[levels-1].getHeight();

    if (ROI)
    {
        xminsearch=roi_lx/((int)pow(2.0,(double)(levels-1)));
        yminsearch=roi_ly/((int)pow(2.0,(double)(levels-1)));
        xmaxsearch=roi_ux/((int)pow(2.0,(double)(levels-1)));
        ymaxsearch=roi_uy/((int)pow(2.0,(double)(levels-1)));
    }

    for (int l=levels-1; l>=0; l--)
    {
        //object[l].display();
        //image[l].display();

        cmincorr=FLT_MAX;
        for (int j=yminsearch; j<ymaxsearch-object[l].getHeight(); j++)
            for (int i=xminsearch; i<xmaxsearch-object[l].getWidth(); i++)
            {
                float corr=0.0;
                int count=0;
                for (int b=0; b<object[l].getHeight(); b++)
                    for (int a=0; a<object[l].getWidth(); a++)
                    {
                        if (!chamferModel) if (object[l].pixel(a,b)>127) continue;
                        corr+=(float)abs((int)image[l].pixel(i+a,j+b)-(int)object[l].pixel(a,b));
                        count++;
                    }

                corr/=count;
                if (corr<cmincorr)
                {
                    xmin=i; ymin=j;
                    cmincorr=corr;
                }
            }

        if (l!=0)
        {
            xminsearch=(xmin*2)-1-20;
            yminsearch=(ymin*2)-1-20;

            xmaxsearch=(xmin*2)-1+object[l-1].getWidth()+20;
            ymaxsearch=(ymin*2)-1+object[l-1].getHeight()+20;

            if (xminsearch<0) xminsearch=0;
            if (yminsearch<0) yminsearch=0;

            if (xmaxsearch>image[l-1].getWidth()) xmaxsearch=image[l-1].getWidth();
            if (ymaxsearch>image[l-1].getHeight()) ymaxsearch=image[l-1].getHeight();
        }


    }

    x=xmin;
    y=ymin;
    corrval=cmincorr;

}



#ifdef HAVE_XALAN
void chamfer_match::saveChamferParameters( const char *fileName )
  throw (mimasexception)
{
    // To be tested.
    ofstream ostream( fileName );

    ostream << "<?xml version=\"1.0\"?>" << endl
    << "<!DOCTYPE object_parameters SYSTEM \"object_parameters.dtd\">"
    << endl
    << "<object_parameters>" << endl
    << "<mcm_recognition_parameters>" << endl;

    storeChamferParams( ostream );

    ostream << "</mcm_recognition_parameters>" << endl;
    ostream << "</object_parameters>" << endl;
}

void chamfer_match::loadChamferParameters( const char *fileName )
  throw (mimasexception)
{
    // Removed check on node-name (must be "object_parameters").
    // XSD-file to be provided.
    xml_document domDocument( "" );
    domDocument.load( fileName );
    xml_reference< XalanElement > docElement
      ( domDocument.getDocumentElement() );
    extractChamferParams( docElement );
}

void chamfer_match::storeChamferParams( ostream &stream )
{
  stream << "<roi "
    << "roi_lx=\"" << roi_lx << "\" "
    << "roi_ly=\"" << roi_ly << "\" "
    << "roi_ux=\"" << roi_ux << "\" "
    << "roi_uy=\"" << roi_uy << "\" "
    << "/>" << endl
    << "<canny_mimas "
    << "standard_dev=\"" << cannyStdDev << "\" "
    << "low_thres=\"" << cannyLowThres << "\" "
    << "high_thres=\"" << cannyHighThres << "\" "
    << "auto_thres=\"" << (int)cannyAutoThres << "\" "
    << "/>" << endl
    << "<chamfer_match "
    << "angle1_search_rad=\"" << ang1 << "\" "
    << "angle2_search_rad=\"" << ang2 << "\" "
    << "anglestep_search_rad=\"" << angstep << "\" "
    << "multiresolution_levels=\"" << mlevels << "\" "
    << "chamfer_model=\"" << (int)chamferModel << "\" "
    << "use_roi=\"" << (int)ROI << "\" "
    << "/>" << endl
    << "<trackingroi "
    << "track_roi_lx=\"" << track_roi_lx << "\" "
    << "track_roi_ly=\"" << track_roi_ly << "\" "
    << "track_roi_ux=\"" << track_roi_ux << "\" "
    << "track_roi_uy=\"" << track_roi_uy << "\" "
    << "/>" << endl
    << "<other "
    << "modelname=\"\" "
    << "/>" << endl;
};

void chamfer_match::extractChamferParams
  ( xml_reference< XalanElement > &docElement ) throw (mimasexception)
{
    xml_reference< XalanElement > n
      ( dynamic_pointer_cast< XalanElement >
        ( docElement.selectNode( "mcm_recognition_parameters" ) ) );

    roi_lx = atoi( n.selectNode( "roi/@roi_lx" ).getNodeText().c_str() );
    roi_ly = atoi( n.selectNode( "roi/@roi_ly" ).getNodeText().c_str() );
    roi_ux = atoi( n.selectNode( "roi/@roi_ux" ).getNodeText().c_str() );
    roi_uy = atoi( n.selectNode( "roi/@roi_uy" ).getNodeText().c_str() );

    cannyStdDev = atof( n.selectNode( "canny_mimas/@standard_dev" ).
                        getNodeText().c_str() );
    cannyLowThres = atoi( n.selectNode( "canny_mimas/@low_thres" ).
                          getNodeText().c_str() );
    cannyHighThres = atoi( n.selectNode( "canny_mimas/@high_thres" ).
                           getNodeText().c_str() );
    cannyAutoThres = atoi( n.selectNode( "canny_mimas/@auto_thres" ).
                           getNodeText().c_str() );

    ang1 = atof( n.selectNode( "chamfer_match/@angle1_search_rad" ).
                 getNodeText().c_str() );
    ang2 = atof( n.selectNode( "chamfer_match/@angle2_search_rad" ).
                 getNodeText().c_str() );
    angstep = atof( n.selectNode( "chamfer_match/@anglestep_search_rad" ).
                    getNodeText().c_str() );
    mlevels = atoi( n.selectNode( "chamfer_match/@multiresolution_levels" ).
                    getNodeText().c_str() );
    chamferModel = atoi( n.selectNode( "chamfer_match/@chamfer_model" ).
                         getNodeText().c_str() );
    ROI = atoi( n.selectNode( "chamfer_match/@use_roi" ).
                getNodeText().c_str() ) != 0;

    track_roi_lx = atoi( n.selectNode( "trackingroi/@track_roi_lx" ).
                         getNodeText().c_str() );
    track_roi_ly = atoi( n.selectNode( "trackingroi/@track_roi_ly" ).
                         getNodeText().c_str() );
    track_roi_ux = atoi( n.selectNode( "trackingroi/@track_roi_ux" ).
                         getNodeText().c_str() );
    track_roi_uy = atoi( n.selectNode( "trackingroi/@track_roi_uy" ).
                         getNodeText().c_str() );

    setName( n.selectNode( "other/@modelname" ).getNodeText() );
}
#endif

}
