#include <boost/shared_array.hpp>
#include "colourspace.h"
#include "image_xineinput.h"

#ifdef HAVE_XINE

using namespace boost;
using namespace std;

namespace mimas {

template<>
void image_xineinput< rgba< unsigned char > >::read
  ( image< rgba< unsigned char > > &img ) throw (mimasexception)
{
  // Get size and format of capture.
  if ( width == -1 || height == -1 ) {
    int ratio_code = 0;
    xine_get_current_frame( stream, &width, &height,
                            &ratio_code, &format, NULL );
  };

#ifndef NDEBUG
  cerr << "xine can give frame of size " << width << 'x' << height << endl;
#endif

  MMERROR( width > 0 && height > 0, mimasexception, ,
           "xine failed to decode frame." );

  switch ( format ) {
  case XINE_IMGFMT_YV12:
    {
      shared_array< unsigned char > yv12
	( new unsigned char[ width * height * 3 / 2 ] );
      img.init( width, height );
      int ratio_code;
      MMERROR( xine_get_current_frame( stream, &width, &height, &ratio_code,
                                       &format, (uint8_t *)yv12.get() ),
	       mimasexception, , "xine failed to capture the YV12 frame" );
      yv12_to_rgba( (const char *)yv12.get(), width, height,
                    (char *)img.rawData() );
    };
    break;
  case XINE_IMGFMT_YUY2:
    {
      shared_array< unsigned char > yuy2
	( new unsigned char[ width * height * 2 ] );
      img.init( width, height );
      int ratio_code;
      MMERROR( xine_get_current_frame( stream, &width, &height, &ratio_code,
                                       &format, (uint8_t *)yuy2.get() ),
	       mimasexception, , "xine failed to capture the YUY2 frame" );
      yuy2_to_rgba( (const char *)yuy2.get(), width, height,
                    (char *)img.rawData() );
    };
    break;
  default:
    // According to the current xine documentation, this should never happen.
    MMERROR( false, mimasexception, ,
	     "Mimas does not support capture for "
	     << (char)( format & 0x000000FF )
	     << (char)( ( format & 0x0000FF00 ) >> 8 )
	     << (char)( ( format & 0x00FF0000 ) >> 16 )
	     << (char)( ( format & 0xFF000000 ) >> 24 )
	     << "-format" );
  };



  // Update state.
  int stream_status = xine_get_status( stream );
#ifndef NDEBUG
  cerr << "status of xine stream is " << stream_status << endl;
#endif
  state = stream_status == XINE_STATUS_PLAY;
}

template<>
void image_xineinput< unsigned char >::read
  ( image< unsigned char > &img ) throw (mimasexception)
{
  // Get size and format of capture.
  if ( width == -1 || height== -1 ) {
    int ratio_code = 0;
    xine_get_current_frame( stream, &width, &height,
                            &ratio_code, &format, NULL );
  };

#ifndef NDEBUG
  cerr << "xine can give frame of size " << width << 'x' << height << endl;
#endif

  MMERROR( width > 0 && height > 0, mimasexception, ,
           "xine failed to decode frame." );
	      
  switch ( format ) {
  case XINE_IMGFMT_YV12:
    {
      shared_array< unsigned char > yv12
        ( new unsigned char[ width * height * 3 / 2 ] );
      img.init( width, height );
      int ratio_code;
      MMERROR( xine_get_current_frame( stream, &width, &height, &ratio_code,
                                       &format, (uint8_t *)yv12.get() ),
	       mimasexception, , "xine failed to capture the YV12 frame" );
      yv12_to_grey( (const char *)yv12.get(), width, height,
                    (char *)img.rawData() );
    };
    break;
  case XINE_IMGFMT_YUY2:
    {
      shared_array< unsigned char > yuy2
	( new unsigned char[ width * height * 2 ] );
      img.init( width, height );
      int ratio_code;
      MMERROR( xine_get_current_frame( stream, &width, &height, &ratio_code,
                                       &format, (uint8_t *)yuy2.get() ),
	       mimasexception, , "xine failed to capture the YUY2 frame" );
      yuy2_to_grey( (const char *)yuy2.get(), width, height,
                    (char *)img.rawData() );
                    };
    break;
  default:
    // According to the current xine documentation, this should never happen.
    MMERROR( false, mimasexception, ,
	     "Mimas does not support capture for "
	     << (char)( format & 0x000000FF )
	     << (char)( ( format & 0x0000FF00 ) >> 8 )
	     << (char)( ( format & 0x00FF0000 ) >> 16 )
	     << (char)( ( format & 0xFF000000 ) >> 24 )
	     << "-format" );
  };



  // Update state.
  int stream_status = xine_get_status( stream );
#ifndef NDEBUG
  cerr << "status of xine stream is " << stream_status << endl;
#endif
  state = stream_status == XINE_STATUS_PLAY;
}

};

#endif
